#include <opencv2/opencv.hpp>
#include <iostream>
#include <fstream>

void readFeatures(std::string fn, std::vector<cv::Point2f>& p1, std::vector<cv::Point2f>& p2)
{
    std::ifstream f(fn);
    if (!f.is_open())
    {
        std::cout << "Couldn't read features file!" << std::endl;
    }

    float x1, y1, x2, y2;
    while (!f.eof())
    {
        f >> x1 >> y1 >> x2 >> y2;
        p1.push_back({ x1, y1 });
        p2.push_back({ x2, y2 });
    }

    f.close();
}

cv::Mat calcHomography(const std::vector<cv::Point2f>& p1, const std::vector<cv::Point2f>& p2)
{
    cv::Mat A(2 * p1.size(), 9, CV_32F);
    for (int i = 0; i < p1.size(); ++i)
    {
        float u1 = p1[i].x;
        float v1 = p1[i].y;

        float u2 = p2[i].x;
        float v2 = p2[i].y;

        A.at<float>(2 * i, 0) = u1;
        A.at<float>(2 * i, 1) = v1;
        A.at<float>(2 * i, 2) = 1.0;
        A.at<float>(2 * i, 3) = 0.0;
        A.at<float>(2 * i, 4) = 0.0;
        A.at<float>(2 * i, 5) = 0.0;
        A.at<float>(2 * i, 6) = -u1*u2;
        A.at<float>(2 * i, 7) = -v1*u2;
        A.at<float>(2 * i, 8) = -u2;

        A.at<float>(2 * i + 1, 0) = 0.0;
        A.at<float>(2 * i + 1, 1) = 0.0;
        A.at<float>(2 * i + 1, 2) = 0.0;
        A.at<float>(2 * i + 1, 3) = u1;
        A.at<float>(2 * i + 1, 4) = v1;
        A.at<float>(2 * i + 1, 5) = 1.0;
        A.at<float>(2 * i + 1, 6) = -u1*v2;
        A.at<float>(2 * i + 1, 7) = -v1*v2;
        A.at<float>(2 * i + 1, 8) = -v2;
    }

    cv::Mat evals, evecs;
    cv::eigen(A.t() * A, evals, evecs);

    cv::Mat H(3, 3, CV_32F);
    for (int i = 0; i < 9; ++i)
    {
        H.at<float>(i / 3, i % 3) = evecs.at<float>(8, i);
    }


    H = H * (1.0 / H.at<float>(2, 2));
    return H;
}

int main()
{
    cv::Mat img1 = cv::imread("a.png");
    cv::Mat img2 = cv::imread("b.png");
    
    std::vector<cv::Point2f> p1;
    std::vector<cv::Point2f> p2;

    std::vector<cv::KeyPoint> kp1;
    std::vector<cv::KeyPoint> kp2;
    std::vector<cv::DMatch> matches;

    // (a) Reading features from file
    //readFeatures("features.txt", p1, p2);
     
    // --  Creating cv::KeyPoints and cv::DMatch vector for visualization
    //for (int i = 0; i < p1.size(); ++i)
    //{
    //    kp1.push_back(cv::KeyPoint(p1[i].x, p1[i].y, 1));
    //    kp2.push_back(cv::KeyPoint(p2[i].x, p2[i].y, 1));
    //    matches.push_back(cv::DMatch(i, i, 0));
    //}

    // (b) Using OpenCV implemented feature detector and feature matcher
    cv::Ptr<cv::SIFT> detector = cv::SIFT::create();
    cv::Mat descript1, descript2;
    detector->detectAndCompute(img1, cv::noArray(), kp1, descript1);
    detector->detectAndCompute(img2, cv::noArray(), kp2, descript2);
    
    cv::FlannBasedMatcher matcher;
    matcher.match(descript1, descript2, matches);

    // --  Convert from cv::KeyPoint to cv::Point2f based on the found matches
    for (int i = 0; i < matches.size(); ++i)
    {
        p1.push_back(kp1[matches[i].queryIdx].pt);
        p2.push_back(kp2[matches[i].trainIdx].pt);
    }


    // Visualization of detected features
    cv::Mat newImg;
    cv::drawMatches(img1, kp1, img2, kp2, matches, newImg);
    cv::imshow("Features", newImg);
    cv::waitKey(0);

    
    cv::Mat H(3, 3, CV_32F);
    // (a) Own implementation of homography estimation
    //H = calcHomography(p1, p2);
    // (b) OpenCV built-in homography estimator
    H = cv::findHomography(p1, p2 , cv::noArray(), cv::RANSAC);
    std::cout << H << std::endl;

    // Warp images based on H
    cv::warpPerspective(img2, newImg, cv::Mat::eye(3, 3, CV_32F), newImg.size());
    cv::warpPerspective(img1, newImg, H, newImg.size(), 1, cv::BORDER_TRANSPARENT);
    
    cv::imshow("Homography", newImg);
    cv::waitKey(0);

    return 0;
}

